# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""This module provides a client for making requests to the `Fidor API`_.

.. _Fidor API: https://developer.fidor.de/

Example usage: ::

    from fidor.client import get_client, initialize_client

    # Initialize a Fidor API client
    client = initialize_client('access token', sandbox=True)

    # Get the current user
    user = client.user.current()

    # Get the account with id 123
    account = client.account.self(variables={'id': '123'})

    # Get 50 transactions
    transactions = client.transaction.instances(params={'per_page': 50})

    # Create a SEPA direct debit
    data = {
        'sepa_direct_debit': {
            'account_id': '123',
            'amount': 1000,
            'collection_date': '2016-12-31',
            'external_uid': 'foobar',
            'mandate_id': '456',
        }
    }
    debit = client.sepa_direct_debit.create(json=data)

    # Access the client instance after initialization
    singleton = get_client()
    assert singleton is client
"""

from __future__ import unicode_literals

from functools import partial

import requests

from six.moves.urllib.parse import urlunsplit

from uritemplate import expand

from fidor import __version__, schemas
from fidor.exceptions import FidorAPIError, LinkNotFoundError, SchemaNotFoundError


class Requester(object):
    """A class which carries out HTTP requests to the Fidor API.

    The link relations of the specified schema are exposed as methods. In
    other words, the "rel" property's value of each Link Description Object
    present in the schema is the name of a method for making requests to
    the related endpoint.

    :param client: A :class:`FidorClient` instance
    :param string schema_name: The name of a schema that defines link relations
    """

    def __init__(self, client, schema_name):
        self.client = client
        self.schema_name = schema_name

    def find_http_method(self, relation_name):
        """Return the HTTP method to use to invoke a Fidor API endpoint.

        :param string relation_name: The name of the relation ("rel" property)
        :raises SchemaNotFoundError: if the schema doesn't exist
        :raises LinkNotFoundError: if the link relation doesn't exist
        """
        link = schemas.get_link(self.schema_name, relation_name)
        method = link.get('method', 'GET')

        return method.upper()

    def make_url(self, relation_name, variables=None):
        """Return the absolute URL of a Fidor API endpoint.

        :param string relation_name: The name of the relation ("rel" property)
        :param dict variables: The variables to use for expanding the URI Template
        :raises SchemaNotFoundError: if the schema doesn't exist
        :raises LinkNotFoundError: if the link relation doesn't exist
        """
        link = schemas.get_link(self.schema_name, relation_name)
        path = link['href']

        if variables is None:
            variables = {}

        path = expand(path, variables)
        netloc = 'aps.fidor.de' if self.client.sandbox else 'api.fidor.de'
        parts = ('https', netloc, path, '', '')

        return urlunsplit(parts)

    def request(self, relation_name, variables=None, params=None, json=None):
        """Send a request and return a JSON-decoded response.

        :param string relation_name: The name of the relation ("rel" property)
        :param dict variables: The variables to use for expanding the URI Template
        :param dict params: The query string parameters
        :param dict json: JSON data to send in the body
        :raises FidorAPIError: if the response has an HTTP 4xx or 5xx status code
        :raises SchemaNotFoundError: if the schema doesn't exist
        :raises LinkNotFoundError: if the link relation doesn't exist
        """
        method = self.find_http_method(relation_name)
        url = self.make_url(relation_name, variables)

        with self.client.session:
            response = self.client.session.request(
                method=method,
                url=url,
                params=params,
                json=json,
                verify=True,
            )

        if not response.ok:
            raise FidorAPIError(response)

        return response.json(use_decimal=True)

    def __getattr__(self, name):
        try:
            schemas.get_link(self.schema_name, name)
        except (SchemaNotFoundError, LinkNotFoundError):
            message = "'{}' schema has no link relation '{}'.".format(self.schema_name, name)
            raise AttributeError(message)

        return partial(self.request, name)


class FidorClient(object):
    """A client for accessing the Fidor API.

    The Fidor API JSON schemas are exposed as attributes. In other words, the
    "name" keyword's value of each schema that contains link relations (Link
    Description Objects) is the name of an attribute, which is an instance
    of the :class:`Requester` class.

    :param string access_token: The access token that grants access to the Fidor API
    :param bool sandbox: If `True` use the sandbox, otherwise the production environment
    """

    def __init__(self, access_token, sandbox=True):
        headers = {
            'Accept': 'application/json',
            'Accept-Charset': 'utf-8',
            'User-Agent': 'fidorpy/{}'.format(__version__),
        }

        self.session = requests.Session()
        self.session.params['access_token'] = access_token
        self.session.headers.update(headers)

        self.sandbox = sandbox

    def __getattr__(self, name):
        try:
            links = schemas.find_links(name)
        except SchemaNotFoundError:
            message = "'{}' schema doesn't exist.".format(name)
            raise AttributeError(message)

        if not links:
            message = "'{}' schema has no link relations.".format(name)
            raise AttributeError(message)

        return Requester(self, name)


_FIDOR_CLIENT = None


def initialize_client(access_token=None, sandbox=True):
    """Initialize and return a Fidor API client with the given arguments.

    :param string access_token: The access token that grants access to the Fidor API
    :param bool sandbox: If `True` use the sandbox, otherwise the production environment
    """
    global _FIDOR_CLIENT
    _FIDOR_CLIENT = FidorClient(access_token, sandbox) if access_token is not None else None
    return _FIDOR_CLIENT


def get_client():
    """Return the previously initialized Fidor API client."""
    return _FIDOR_CLIENT
