# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.

"""This module allows easy access to the `Fidor API JSON Schema`_. The release
version of the schemas bundled with fidorpy is specified by the value of this
module's ``__version__`` attribute.

.. _Fidor API JSON Schema: https://github.com/fidor/fidor_schema
"""

from __future__ import unicode_literals

import os

from fidor.exceptions import LinkNotFoundError, SchemaNotFoundError
from fidor.utils import load_json_file


__version__ = '0.4.2'


def find_schemas():
    """For each Fidor API JSON schema found, yield a 3-tuple containing
    the schema's name, its absolute file system path and the schema
    itself as a dictionary.
    """
    here = os.path.dirname(os.path.abspath(__file__))
    json_dir = os.path.join(here, 'json')

    for root, _, files in os.walk(json_dir):
        for filename in files:
            path = os.path.join(root, filename)
            schema = load_json_file(path)
            try:
                name = schema['name']
            except KeyError:
                pass
            else:
                yield name, path, schema


SCHEMAS = {name: schema for name, _, schema in find_schemas()}


def find_links(schema_name):
    """Return a dictionary mapping the "rel" property's value of each Link
    Description Object (LDO) present in the specified schema to the LDO
    itself. If no LDOs are present, an empty dictionary is returned.

    :param string schema_name: The name of the JSON schema
    :raises SchemaNotFoundError: if the schema doesn't exist
    """
    try:
        schema = SCHEMAS[schema_name]
    except KeyError:
        message = "'{}' doesn't exist.".format(schema_name)
        raise SchemaNotFoundError(message)

    links = schema.get('links', [])
    return {link['rel']: link for link in links}


def get_link(schema_name, relation_name):
    """Return the Link Description Object defined in the specified schema
    and identified by the relation's name ("rel" property).

    :param string schema_name: The name of the JSON schema
    :param string relation_name: The name of the relation ("rel" property)
    :raises SchemaNotFoundError: if the schema doesn't exist
    :raises LinkNotFoundError: if the link doesn't exist
    """
    links = find_links(schema_name)

    try:
        ldo = links[relation_name]
    except KeyError:
        message = "'{}' doesn't exist.".format(relation_name)
        raise LinkNotFoundError(message)

    return ldo
